/*******************************************************************************
        An example of a binary tree problem

The horse winning problem disccused on pp. 133-136 of Clocksin & Mellish,
third edition.
*******************************************************************************/
% For SWI Prolog need the following.

:- dynamic(db/1).
:- dynamic(db1/1).

/*******************************************************************************
The first set of rules define a "<" relationship for sets of characters.
The relationship is based on the ASCII ordering.

aless(X,Y)  Asserts that X is less than Y.
name(A,B)   Asserts that A is the internal representation of an atom and B is
            its corresponding set of characters making up the atom.

alessx(X,Y) Asserts the character set X is less than the character set Y.
*******************************************************************************/

aless(X,Y) :- name(X,L) , name(Y,M) , alessx(L,M).

alessx([],[_|_]).
alessx([X|_],[Y|_]) :- X < Y.
alessx([A|X],[B|Y]) :- A = B , alessx(X,Y).

/*******************************************************************************
The binary tree is defined as a compound term (see EngFreDictionary.pl).
The traditional binary tree structure is represented as a string of characters.
Each branch, instead of being a pointer, is the string representation of the
corresponding sub-tree. The null subtree is represented as the "I don't care"
variable. Our tree has two data values: the name of the horse and the value of
its winnings.  The typical node strtucture is:

  w( horse''sName , winnings , leftSubTree , rightSubTree )
*******************************************************************************/

:- assertz(db(w(massinga,858,
              w(braemar,385, w(adela,588,_,_),_),
  						w(panorama,158, w(nettleweed,579,_,_),_)))).

:- assertz(db1(w(massinga,858,_,_))).

/*******************************************************************************
The look-up rule is defined below.  The first rule uses the device of accessing
the database to find the binary tree to use.
*******************************************************************************/

lookup(H,G) :- db(BT) , lookup(H,BT,G)
%             , write(BT) , nl 
.

lookup1(H,G) :- db1(BT) , lookup(H,BT,G)
%              , write(BT) , nl
.

/** 
The remaining rules define a binary search on a binary tree.
**/

lookup(H,w(H,G,_,_),G) :- !.

lookup(H,w(H1,_,Before,_),G) :- aless(H,H1) , lookup(H,Before,G).

lookup(H,w(H1,_,_,After),G)  :- aless(H1,H) , lookup(H,After,G).

/*******************************************************************************
Since the left and right tree are uninstantiated variables in base rule, lookup
has a side effect of adding an entry into the binary tree, if one cannot be found.

  !!! Adding to the tree is equivalent to
      looking up a missing entry !!!

add1 takes advantage of this to save the old database and insert a new fact
into the database.
*******************************************************************************/

add1(H,G) :- db1(BT) , lookup(H,BT,G) , assertz(db1(BT)).

/*******************************************************************************
Now both binary trees are in the database.
Try db(BT) ; <return> to see how many binary trees are returned.

So we use a retract to remove the old binary tree from the database
and insert the new binary tree into the database.
*******************************************************************************/

add2(H,G) :- db1(BT) , OldBT = BT,
             lookup(H,BT,G),
             retract(db1(OldBT)),
             assertz(db1(BT)).

/*******************************************************************************
We do not actually need to save the old binary tree, since BT is the tree.
*******************************************************************************/

add3(H,G) :- db1(BT) , retract(db1(BT)),
             lookup(H,BT,G) , assertz(db1(BT)).

add4(H,G) :- db(BT) , retract(db(BT)),
             lookup(H,BT,G) , assertz(db(BT)).

/*******************************************************************************
Deleting from a binary tree.
deltree(Tree, H, G, New_tree) -- deleting node with H in it is the new tree
*******************************************************************************/

% Base cases delTree occur when node is found.

deltree(w(H,G,L,R), H, G, R) 						 :- var(L) , !.
deltree(w(H,G,L,R), H, G, L) 						 :- var(R) , !.
deltree(w(H,G,L,R), H, G, w(H1,G1,L,R1)) :- delmin(R,H1,G1,R1).

% Recursive cases search the tree.

deltree(w(H1,G1,L,R), H, G, w(H1,G1,L1,R)) :-
     aless(H,H1) , deltree(L,H,G,L1).

deltree(w(H1,G1,L,R), H, G, w(H1,G1,L,R1)) :-
     aless(H1,H) , deltree(R,H,G,R1).

/* delmin(Tree, H, G, New_tree) searches the right sub-tree
for the left most node; the nearest right neighbour.

deleting minimum item H in Tree is New_tree.
Return the data values for the deleted item through H and G.
*/

delmin(w(H,G,L,R), H, G, R) :- var(L) , !.

delmin(w(H1,G1,L,R), H, G, w(H1,G1,L1,R)) :- delmin(L,H,G,L1).

/*******************************************************************************
add(Tree, D, New_tree) -- adding D to the Tree is the New_tree. Has the property
that deletion is done by reversing which is the variable. add(New_tree, D, Tree)
- Tree an instance, then New_tree is Tree without D. This is a non-deterministic
algorithm, as insertion can occur at any level.

Example query:
  add(nil,3,T1), add(T1,4,T2)
, add(T2,1,T3), add(T3,2,T4), add(FT,3,T4).

3, 4, 1 and 2 are added to the tree, then 3 is removed.
FT is the tree containing 4, 1 and 2.

Use show(T) after forming T4.
*******************************************************************************/

% Add D as a new root

add(Tree, D, New_tree) :- addroot(Tree, D, New_tree).

% Add D into left sub-tree

add(t(L,D1,R), D, t(L1,D1,R)) :- D < D1,
 %, write(D) , write(' at 1: ')
                                 add(L, D, L1).

% Add D into right sub-tree

add(t(L,D1,R), D, t(L,D1,R1)) :- D1 < D,
 % , write(D) , write(' at 2: ')
                                 add(R, D, R1).

% Insert into empty tree.

addroot(nil, D, t(nil,D,nil)).

% Insert into left sub-tree

addroot(t(L,D1,R), D, t(L1,D,t(L2,D1,R))) :- D < D1 , addroot(L, D, t(L1,D,L2)).

% Insert into right sub-tee

addroot(t(L,D1,R), D, t(t(L,D1,R1), D, R2)) :- D1 < D , addroot(R, D, t(R1,D,R2)).

/*******************************************************************************
Show the tree structure
*******************************************************************************/

show(Tree) :- nl , show(Tree,0) , nl.
show(nil, _).
show(t(Left, D, Right), Indent) :-
    Indent2 is Indent+2,
    show(Right,Indent2),
    tab(Indent), write(D), nl,
    show(Left, Indent2).

showh(Tree) :- nl , showh(Tree,0) , nl.
showh(Tree, _) :- var(Tree).
showh(w(H, _, Left, Right), Indent) :-
    Indent2 is Indent+2,
    showh(Right,Indent2),
    tab(Indent), write(H), nl,
    showh(Left, Indent2).
