/*******************************************************************************
         Parse and response rules for the examples discussed in class.

/*******************************************************************************
The parsing rules are given below. The first rule is in standard Prolog to
handle the special case of stopping the main loop of the program. The remaining
rules make use of the grammar rule syntax built in to Prolog. Note that the
grammar rule syntax is itself defined within Prolog using 'standard' Prolog. 

The parse rules 'understand' the following sentence structures.

1. Stop.               Terminate chatting.
2. ____ is a ____.     Example:  Mary is a woman.
3. A ____ is a ____.   Example:  A woman is a person.
4. Is ____ a ____?     Example:  Is Mary a person?
*******************************************************************************/

:- consult('chatBasis.pl').

/******
parse(Clause, Sentence, Left-over)
-- Clause is the encoded result of parsing the input sentence.
-- Sentence is a list that matches the prefix of the input.
-- Left_over is what is permitted after the prefix has been parsed.

The first rule, in Prolog, produces the result 'stop' if the input matches 'Stop.'
with nothing left over.  This rule is used to terminate the chat program.
******/

parse(stop, [ 'Stop', '.' ], []) .

/******
The second, third and fourth rules are a clauses in a "Definitive Clause
Grammar" that are translated into standard Prolog.

The second rule matches sentences of the type  Mary is a woman.  The result
Clause is the predicate  T(Name)  formed from the values of T and Name. 
******/
 
parse(Clause) --> thing(Name) , [ is , a ] , type(T) , ['.']
                , { Clause =.. [ T , Name ] , ! } .

/******
The third rule matches sentences of the type  A woman is a person.  The result
Clause is the rule  T2(X) :- T1(X)  formed from the values of T1, T2 and the
variable X.  
******/
 
parse(Clause) --> ['A'] , type(T1) , [ is , a ] , type(T2) , ['.']
                , { Head =.. [T2, X] , Condition =.. [ T1, X ]
                  , Clause = (Head :- Condition) ,! } .

/******
The fourth rule matches sentences of the type  Is Mary a person?  The result
Clause is the structure  ?(T(Name))  formed from the values of T and Name.
******/
 
parse(Clause) --> ['Is'] , thing(Name) , [ a ] , type(T) , ['?']
                , { Goal =.. [T, Name] , Clause = ('?'(Goal)) , ! } .

/******
The last rule, in Prolog, is the catch all rule indicating that none of the
previous parse rules succeeded in matching the input.  The result is 'noparse'
independent of the value of Sentence and Left_over.
******/

parse(noparse, _, _) .
 
/*******************************************************************************
                  SplitOff rules for parsing.
In the example period, '.', comma, ',' and possesive, "'s" are split off the
end of input words.
*******************************************************************************/

   splitOff("'s") .
   splitOff("?") .
   splitOff(".") .

/*******************************************************************************
Set of rules to notify the user of what the parse rules did.
*******************************************************************************/

/******
If the result of parsing is the structure  ?(Goal) and if the query Goal
succeeds, then write the sentence  Yes A1 is a F.  Otherwise write No to
indicate the Goal failed.
******/

respondTo('?'(Goal)) :- Goal =.. [F,A1] ,
                       ( Goal -> write('Yes, ') , write(A1) , write(' is a '),
                                 write(F), write('.')
                         ; write('No') )
                       , ! , nl , nl.

/******
If the result of parsing is noparse, then inform the user that the input was
not parsable.
******/
 
respondTo(noparse) :- write('Can''t parse that.' ) , nl , ! .

/******
If the result of parsing is stop, then tell the user the program is terminating.
******/
 
respondTo(stop) :- write('All done.') , nl , ! .

/******
The last rule is the catch all for the example that inserts the clause at the
beginning of the database and informs the user.
******/
 
respondTo(Clause) :- asserta(Clause) , write('Ok') , nl , ! .
