/*   An example creating a tiny English-French dictionary

The example includes using a "simulated" list structure.  The list structure
given in Prolog (using the [...] notation) is syntactic sugar for such a list
structure
******************************************************************************
First we define a membership relation for a list where a list is the structure
    list(H,T)  H is the first element on the list, T is the rest of the list.
******************************************************************************/

memberSimulated(X, list(X,_)).
memberSimulated(X, list(_,L)) :- 
             memberSimulated(X,L).

/******************************************************************************
We can define both an English-French dictionary and a look-up relation as a 
single rule.
     french(E,F)  where E and F are "equivalent" English and French words
                  respectively.
The look-up is accomplished by trying to show list membership of a dictionary
"entry" --- entry(E,F) represents on entry in the dictionary.  The second
argument of "member" is the list of entries in the dictionary.  The last
"entry" is "empty" to fill the last component of "list".  The term "empty"
could be any term but we chose "empty" because it is mnemonically significant.
******************************************************************************/

englishFrench1(English,French) :-
	memberSimulated(entry(English,French), 
    list(entry(book,livre),
			 list(entry(man,homme),
			 list(entry(apple,pomme),
				empty))) ).

 
/******************************************************************************
 Let's use Prolog's list structure to accomplish the same task.  The syntactic
sugar makes defining and using list structures more convenient.
******************************************************************************/

englishFrench2(English,French) :-
 	member(entry(English,French),
    [ entry(book,livre),
      entry(man,homme),
      entry(apple,pomme),
	     empty] ).

/******************************************************************************
Having a dictionary in the look-up rule can be inconvenient. So, we define
the dictionary as a fact and modify the rule to find a dictionary.  Now we can
change the dictionary without changing the rule.
******************************************************************************/

dictionary(Name,D) :-
  Name = d1, D = [ entry(book,livre),
                   entry(man,homme),
                   entry(apple,pomme) ]  ;

  Name = d2 , D = [ entry(book,koob),
                    entry(man,nam),
                    entry(apple,elppa)].

englishFrench3(E,F,N) :-
        dictionary(N,D),
        member(entry(E,F),D).

/******************************************************************************
This is also an opportune moment to show how one can define functors to be
infix operators (they can also be prefix and postfix without the use of 
parenthesis).

The operator "op" is used to define a new "operator" in Prolog.  A detailed
description is in Section 3.3 of Bratko, fourth edition.

We define "e" (it looks a bit like epsilon which is used in mathematics to
mean "belongs to".  The operator is given a medium level precedence and is
made to be right associative.

The basic dictionary look-up is shown below.  After the "op" definition for
"belongs to" is the member rule using "e".
******************************************************************************/

  :-  op(500,xfy,[e]).

X e [X|_].
X e [_|L] :- X e L.

englishFrench4(E,F) :-
  entry(E,F) e [ entry(book,livre),
                 entry(man,homme),
                 entry(apple,pomme) ].

:-  op(500,xfy,[belongs_to]).

X belongs_to [X|_].
X belongs_to [_|L] :- X belongs_to L.

englishFrench5(E,F) :-
  entry(E,F) belongs_to [ entry(book,livre),
                 entry(man,homme),
                 entry(apple,pomme) ].
