/*******************************************************************************
An example file illustrating some aspects of PlUnit testing that you can use.

The first part of the file is the definition of the predicates that will be
tested.

Tests are run by consulting this file and then invoking the predicate
run_tests

Documentation on using PlUnit for testing Prolog programs can be found at
the following URL.
          http://www.swi-prolog.org/pldoc/package/plunit.html
*******************************************************************************/

:- consult('tower.pl').
:- consult('ring.pl').

/* The predicate total asserts that Total is the sum of all the numbers in
the list Integers.

Require: Integers to be a list of numbers
Ensure: Total = +/Integers
*/

total([],0).
total([First | Rest], Total) :- total(Rest,Total_rest)
                              , Total is First + Total_rest.

/*******************************************************************************
Test cases for the predicates.
*******************************************************************************/
:- begin_tests(set1).

/* The first test verifies the empty list of integers returns the total 0.
While the second test demonstrates a test that fails. */
/* Note that the first test ends with a warning that the test succeeded
   with choicepoint, which means that if the test query was done interactively
   Prolog would give you the choice of entering ';' or 'cr', indicating that
   Prolog could continue searching for another answer.
*/

test(test1_noNumbers_total_to_0) :- total([],0).
test(test2_noNumbers_total_to_0) :- total([],1).

/* The following test does the same work as the previous test, except
   the parameter '[fail]' is used to tell Prolog the test is expected to
   fail, so return true when the test fails and false if the test succeeds.
*/

test(test2b_noNumbers_total_to_0, [fail]) :- total([],1).

/* An integer list of length one has a total equal to the the only integer
in the list. */

test(one_integer_is_total) :- total([4], 4).

/* A length many integer list has a total equal to the sum of the integers.
Test 1 succeeds.  Test 2 is expected to fail as the test is incorrect. */

test(test1_many_integer_list_total) :- total([1, 2, 3, -3, -2, -1], 0).
test(test2_many_integer_list_total, [fail]) :- total([1, 2, 3, -3, -2, -1], 1).

/* Blocks a and on the ground are towers. */
/* Note that the first of following two tests ends with a warning that the test
   succeeded with choicepoint, which means that if the test query was done
   interactively Prolog would give you the choice of entering ';' or 'cr',
  indicating that Prolog could continue searching for another answer.
  
  The second test uses the parameter [nondet] to eliminate the choicepoint
  warning message.
*/

test(a_and_c_valid_towers) :- tower(t(a,ground)) , tower(t(c, ground)).
test(a_and_c_valid_towers_2, [nondet]) :- tower(t(a,ground)) , tower(t(c, ground)).

/* Verify one case of a good and safe tower. */

test(good_and_safe_tower, [nondet]) :-
		goodSafeTower(d, 3, 3, 4, t(d, t(b, t(a,ground)))).

/* Verify the list of characters who die in the ring cycle. */

test(who_dies_test1, [nondet]) :- dies(Who) , Who = mime.

    /* Use of findall in "test2" and all in "test3" takes care of
       choicepoints. They can be used if the number of choices is limited. */

test(who_dies_test2) :- findall(Who, dies(Who), Who_dies)
                , Who_dies = [ fafner, hunding, siegfried, siegmund, mime
                             , fasolt, fasolt, gunther, brunnhilde, sieglinde ].
test(who_dies_test3, all(Who = [ fafner, hunding, siegfried, siegmund, mime
                           , fasolt, fasolt, gunther, brunnhilde, sieglinde ]))
        :- dies(Who).

/*
Expect test reverse4 to succeed inspite of the reverse failure because
we indicated we expected failure.
*/

test(reverse4,[fail]) :- reverse([a,b], [b,a,1]).

:- end_tests(set1).
