/*****************************************************************************
   An example which shows how a maze problem can be defined and solved.

First we define the "facts" which are the edges from intersection to
intersection in the maze.
*****************************************************************************/


edge(a, b).
edge(b, c).  edge(c, d).
edge(d, e).  edge(e, f).
edge(b, e).  edge(e, g).

/*****************************************************************************
Now define a "goTo" rule to traverse the maze.

   go0(A, B, T)   T is the path from A to B.
*****************************************************************************/

           /* If we have reached the destination then write the path. */
go0(D, D, P) :- write(P), nl.

/*****************************************************************************
Now define the rule for a path P from S to D.  We try both arrivals and
departures from S --- we permit and search for all paths including those with
loops.  We try to find a next intersection, N, from S on the way to D.

Notice the heavy use of backtracking to search for all possible paths.
The path is printed backwards.  The path is not returned to the user because
there is no definition given to P.  The starting node is not in the path.
*****************************************************************************/

go0(S, D, P) :-
    ( edge(S, N) , write1(S, N)
    ; edge(N, S) , write1(S, N)
    ),
    go0(N, D, [N | P]).

write1(From, To) :- write([going, from, From, to, To]) , nl.

/*****************************************************************************
Want to eliminate looping.  Can be done by checking if the next node is
already in the path.
*****************************************************************************/

go1(D, D, P) :- write(P), nl.
go1(S, D, P) :-
  ( edge(S, N) , write1(S, N)
  ; edge(N, S) , write1(S, N)
  ),
  not(member(N, P)),           % Do not loop
  go1(N, D, [N | P]).

/*****************************************************************************
Not(member) does not work because P is a variable
   on the way down and N can always be a member by assigning [N] to P.
   So not(member(N,P)) is always false because member(M,P) is always true.

Notice that starting node is not in the path in either go0 or go1.

A new predicate legal is add to go2 although it appears to be equivalent to
not(member(X,T)).
*****************************************************************************/

go2(D, D, P) :- write(P), nl.
go2(S, D, P) :- ( edge(S, N) , write1(S, N)
                ; edge(N, S) , write1(S, N)
                ),
                legal(N, P),
                go2(N, D, [N | P]).

legal(_, []).
legal(X,[H | T]) :- X \= H, legal(X, T).

/*****************************************************************************
go3 creates a forward path using append but does not save the result.
Also splits off the logic for getting the next potential node in a path.
*****************************************************************************/

go3(D, D, P) :- write(P), nl.
go3(S, D, P) :- legalnode(S, P, N),
                append(P, [N], LP) , go3(N, D, LP).

legalnode(X, P, Y) :- (edge(X, Y) ; edge(Y, X)),
                      legal(Y, P).

/*****************************************************************************
We need to match the input variable P with a path in order to get a
path returned.  go4 shows the principle.  Note the return of the destination
node in rule 1, and the inclusion of the start node of a path in rule 2.
*****************************************************************************/

go4(D, D, [D]).
go4(S, D, [S | R]) :- (edge(S, N) ; edge(N, S)),
                      go4(N, D, R).

/*****************************************************************************
The paths get ever longer due to looping so put back legalnode.
*****************************************************************************/

go5(D, D, [D]).
go5(S, D, [S | R]) :- legalnode(S, [S | R], N),
                      go5(N,D,R).

/*****************************************************************************
Oops. legalnode does not work correctly as we are not giving it the proper
prefix of the path.  We need to add another parameter (accumulator) to keep track
of the path so far.

go6 creates a backwards path in an accumulator and puts the accumulator
value into P thereby returning a path.  The path is backwards so use
reverse is used to create a forward path.

****************************************************************************/

go6(S, D, P) :- go6(S,   % Start node
                    D,   % Destination node
                    [],  % Accumulator for the path on the way down
                    RP), % The reverse path, filled in when at the destination
                reverse(RP, P).

go6(D, D, AP, [D|AP]).    % At the end the accumulated path is appended to
                          % the destination node.

go6(S, D, AP, P) :- legalnode(S, AP, N),    % N is legal if it is a successor
                                            % to S and is not in the accumulated
                                            % path.
                    go6(N, D, [S | AP], P). % Extend the path S in the
                                            % in the accumulated path.

/*****************************************************************************
go7 uses Holes.  Holes are automatically used in regular recursion.
Computing the value of the rest of the list fills in the hole (P).
Need a separate legal path (accumulator), as the whole path is not available.

*****************************************************************************/

go7(S, D, P) :- go7(S,   % Start node
                    D,   % Destination node
                    P,   % The path, which is the hole to fill in
                    []). % The accumulator for the path on the way down

go7(D, D, [D], _).     % At the destination node, it is the value to put into
                       % The hole.
 
% 3rd argument [S | P] fills the previous hole, with the hole P to be filled in
% on the recursive call.

go7(S, D, [S | P], T) :- S \= D,    % Need a guard to prevent a spurious 2nd soln
                         legalnode(S, T, N), % N is legal if it is a successor
                                             % to S and is not in the accumulated
                                             % path.
     % P is the hole to fill. Extend the path with N in the accumulated path
                         go7(N, D, P, [N | T]).

/*****************************************************************************
go8 creates a forward path and uses append and an accumulator to
keep a forward path (no reverse at the end).  But loses the first node in the path.
*****************************************************************************/

go8(S, D, P) :- go8(S, D, [], P).
go8(D, D, AP, AP).
go8(S, D, AP, P) :- legalnode(S, AP, N),
                    append(AP, [N], LP),
                    go8(N, D, LP, P).

/*****************************************************************************
go9 solves the problem by starting the accumulator with the start node
instead of the empty list.
*****************************************************************************/

go9(S, D, P) :- go9(S, D, [S], P).
go9(D, D, AP, AP).
go9(S, D, AP, P) :- legalnode(S, AP, N),
                    append(AP, [N], LP),
                    go9(N, D, LP, P).
