/*********************************************************************
Bubble sort
**********************************************************************
bubblesort(Unsorted, Sorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
        using the bubble-sort algorithm.
******************************************/

bubblesort(Unsorted, Sorted) :-         % Handle swap case
    swapApair(Unsorted, Swapped), ! ,   % Cut is necessary
    bubblesort(Swapped, Sorted).

bubblesort(Sorted, Sorted).     % No swaps done

swapApair([First, Second | Rest], [Second, First | Rest]) :-
    First > Second.

swapApair([First | Rest], [First | NewRest]) :-
    swapApair(Rest, NewRest).

/*********************************************************************
Binary tree sort
**********************************************************************
binarytreesort(Unsorted. Sorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
        by building a binary tree and then traversing it in inorder.
******************************************/

binarytreesort(Unsorted, Sorted) :-
    buildTree(Unsorted, Tree, _),
    inordera(Tree, Sorted).

buildTree([], Heap, Heap).

buildTree([First | Rest], Heap, Acc) :-
    insertIntoTree(First, Acc),
    buildTree(Rest, Heap, Acc).

insertIntoTree(Item, bt(_, Item, _)) :- !.

insertIntoTree(Item, bt(Before, TreeItem, _)) :-
    Item =< TreeItem,
    insertIntoTree(Item, Before).

insertIntoTree(Item, bt(_, TreeItem, After)) :-
    Item > TreeItem,
    insertIntoTree(Item, After).

inordera(BT,Order) :- inordera(BT,[],Ot), reverse(Ot,Order).
inordera(X,Order,Order) :- var(X) , !.
inordera(bt(L,D,R),Acc,Order) :- inordera(L,Acc,Ol) , inordera(R, [D|Ol],Order).

/*********************************************************************
Insert sort
    Exercise: Use an accumulator.
    Exercise: Use a hole.
**********************************************************************
insertsort(Sorted, Unsorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
        using the insert-sort algorithm with the list being sorted
        at the end instead of the beginning.
******************************************/

insertsort([], []).

insertsort([First | Rest], Sorted) :-
    insertsort(Rest, SortedRest),
    insert(First, SortedRest, Sorted).

insert(Item, [First | Sorted], [First | NewSorted]) :-
    Item > First , ! ,
    insert(Item, Sorted, NewSorted).

insert(Item, Sorted, [Item | Sorted]).


/*********************************************************************
Selection sort
    Exercise: Sort the tail of the list and select from the front part.
    Exercise: Use a hole.
**********************************************************************
sectionsort(Unsorted, Sorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
        using the selction-sort algorithm.
******************************************/

selectionsort(Unsorted, Sorted) :-
    selectionsort(Unsorted, Sorted, []).

selectionsort([], Sorted, Sorted).

selectionsort([First | Rest], Sorted, Acc) :-
    insert(First, Acc, NewAcc),
    selectionsort(Rest, Sorted, NewAcc).
    

/*********************************************************************
Merge sort
**********************************************************************
mergesort(Unsorted, Sorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
        using the merge-sort algorithm.
******************************************/

mergesort(Unsorted, Sorted) :-
    partition(Unsorted, Partitioned, 1),
    mergeAllPairs(Partitioned, Sorted).

/*********************************************************************
Partition a list into sublists of length LengthSublist

partition(List, PartitionedList, LengthSublist)

List: A list of any items

PartitionedList: A list with items I1, I2, ..., In such
    that I1 || || I2 || ... || In = List
    and |Ip| = LengthSublist for p=1..n-1
    and |In| <= LengthSublist

******************************************/

% If the list is shorter than LengthSublist, then all the items in
% the list are the partition.

partition(List, [List], LengthSublist) :-
    length(List, N),
    0 < N , N =< LengthSublist.

% Otherwise remove the first LengthSublist of items from List to form
% the first partition and repeat for the rest of the list.

partition(List, [FirstPartition | RemainingPartitions], LengthSublist)
    :- length(FirstPartition, LengthSublist),  % Proper length sublist
    append(FirstPartition, RestOfList, List),  % Split the List
    partition(RestOfList, RemainingPartitions, LengthSublist).

/*********************************************************************
Merge successive pairs of sublists a list until there is only one
sublist.

mergeAllPairs(List, Result)

List = [NL1, NL2, ..., NLp]
    is a list of lists where each sublist NLi is a list of numbers.

Result = mergeOf(NL1, NL2, ..., NLp)

******************************************/

    % Stop when there is only one list.  Cut is necessary, as a single
    % List will match the recursive case.
mergeAllPairs([List], List) :- length([List],1), ! .

mergeAllPairs(List, Result) :-
    mergePairs(List, RR),
    mergeAllPairs(RR, Result).

/*********************************************************************
Merge successive pairs of a list.

mergePairs(List, MergedList)

List = [NL1, NL2, ..., NLp]
    is a list of lists where each sublist NLi is a list of numbers.

MergedList = [ML1, ML2, ..., MLq]
    where MLi = merge(NLi, NLi+1) for i = 1 .. p/2
    and for odd p, MLq = NLp
******************************************/

mergePairs([], []).                % Length 0 case

mergePairs([Result], [Result]).    % Length 1 case

mergePairs([First, Second | Rest], [FirstMerge | RemainingMerges]) :-
    merge(First, Second, FirstMerge),
    mergePairs(Rest, RemainingMerges).


/*********************************************************************
Quicksort
**********************************************************************
quicksort(Unsorted, Sorted)

Unsorted: List of numbers
Sorted: The numbers in Unsorted sorted in ascending order
    using the quick-sort algorithm.

Quicksort algorithm defined using:
1 - Standard algorothm
2 - Using holes
Page 163 of Clocksin & Mellish
******************************************/

/*
Split(SV, List, Smaller, Bigger)
-- partition List into  < SV and >= SV
*/

split(_, [], [], []).

split(SV, [A | Lr], [A | Sr], Bigger) :-
    A < SV ,
    split(SV, Lr, Sr, Bigger).

split(SV,[A | Lr],Smaller,[A | Br]) :-
    A >= SV ,
    split(SV, Lr, Smaller, Br).

/*****************************************/
/* Quicksort -- qs(Unsorted, Sorted)     */

qs([], []).

qs([SV | T], Sorted) :-
    split(SV, T, Smaller , Bigger) ,
    qs(Smaller, Ss) ,
    qs(Bigger, Bs) ,
    append(Ss, [SV | Bs], Sorted).

/*****************************************/
/* Quicksort with holes                  */

qsh(Unsorted, Sorted) :-
    qsh(Unsorted, Sorted, []).

qsh([], Sorted, Sorted).

qsh([SV | T], Sorted, Hole) :-
    split(SV, T, Smaller, Bigger) ,
    qsh(Smaller, Sorted, [SV | Hole1]) ,
    qsh(Bigger, Hole1, Hole).
