/***************************************************************************
 A simplified block world that recognizes and builds towers.

A tower is represented as a compound term with the following structure.

        t(q , t(w , t(z , t( ... , t(r , ground) ... ))))

where q, w, z and r represent blocks.

First there are a collection of block types.  To build a tower we want to have
the correct types of blocks available and have many instances of a block as
are necessary.
***************************************************************************/

block(a). block(b). block(c). block(d).

/***************************************************************************
Next are facts that tell us that block types a and c can be placed directly
on the ground.
***************************************************************************/

on(a,ground). on(c,ground).

/***************************************************************************
To build towers we need to be able to place multiple blocks on top of each
other.  The following facts state which type of block can be on which other
type of block.  The effect is that towers have an "a" or "c" block as a base
(depending upon ground type or looks).  The middle of a tower is built
of "b" blocks and the top is a "d" block.
***************************************************************************/

on(b,a). on(b,c). on(d,b). on(b,b).

/***************************************************************************
Now we have the rules (clauses) defining a tower.

The base case (no pun intended) is a tower consists of a block that can be on
the ground.

The recursive case is a larger tower can be built out of a smaller tower by
adding a legal block on top.
***************************************************************************/

tower(t(X,Y))       :- block(X) , on(X,Y) , 
                       Y=ground.
tower(t(X,t(Y,Z)))  :- block(X) , on(X,Y) ,
                       tower(t(Y,Z)).

/*  Try
    -- tower(t(a,ground)).
    -- tower(t(d,t(b,t(a,ground)))).
    -- tower(t(b,t(d,t(a,ground)))).

    -- Try tower(T).
    -- What happens? Can you get a tower with a top block d?  Why not?
       What is the remedy?

Create a rule properTower(Topblock,t(...)) that defines a tower such
   that the top block of Tower is of type top.  Add the fact that "d" is
a top block to test your rule.
*/

top(d).
properTower(Tb,t(Tb,t(T,Z))) :- top(Tb),
                                on(Tb,T),
                                tower(t(T,Z)).

/*
Create a set of rules to define a tower with height,
    measuredTower(Height,t(...))
such that Height is the number of blocks in the tower.  You cannot build
on tower as it is of the wrong arity.
*/

measuredTower(H,t(X,Y)) :- block(X) , on(X,Y) ,
                          Y=ground , H=1.

measuredTower(H,t(X,t(Y,Z)))  :- 
    block(X) , 
    on(X,Y) ,
    measuredTower(H1, t(Y,Z)) ,
    H is H1 + 1 .

/*
Create a rule that defines a good and safe tower,
   goodSafeTower(Tb,Height,Min,Max,t(...)), such that it is a proper
tower as in part b and must be between a minimum and maximum height
inclusive.
*/

goodSafeTower(Tb,H,Min,Max,t(T,R)) :-
       properTower(Tb,t(T,R)) ,
       measuredTower(H,t(T,R)) ,
       Min =< H , H =< Max .
/*
If queried with goodSafeTower(Tb,H,3,4,T) the proper towers with heights
between 3 and 4 inclusive are found, and then infinite recursion sets in.
Infinite recursion occurs because the goal Height =< Max fails, backtracking
takes place and new higher proper tower is found, the Max goal fails and so
on.  There is nothing that we can easily do to stop the infinite recursion.
We could test for measuredTower first and put a predetermined limit on
measuredTower so it eventually will fail.
*/